#ifndef bodies_h
#define bodies_h

#include "defo.h"

using namespace std;


enum class NormalizationMethod
// bodies numbered from 0, ..., n-1; last body is computed from the center of mass
{
    #ifdef PLANAR
    equalToZero, // y_{n-2} = 0
    equalToFirst // y_{n-2} = y_0
    #endif
	
	#ifdef SPATIAL
	equalToZero, // z_{0} = 0, y_{n-2} = 0, z_{n-2} = 0
    equalToFirst // not implemented
	#endif
};



/* ========================================================= *
 *   BODIES                                                  *
 *
 * The class encapsulating the configutaion of the bodies,   *
 * along with some getters/setters and helpers.              *
 * ========================================================= */


struct Body
{
    #ifdef PLANAR
    MyInterval x, y;
	#endif
	#ifdef SPATIAL
	MyInterval x, y, z;
	#endif
    MyInterval mass;
};

//  the equation for normalized central configuration can be written as
//  f_{x_i}=x_i + \sum_{j \neq i} m_j \frac{\partial }{\partial x_i} 1/r_{ij}=0
//  f_{y_i}=y_i + \sum_{j \neq i} m_j \frac{\partial }{\partial y_i} 1/r_{ij}=0
// in 3D :  f_{z_i}=z_i + \sum_{j \neq i} m_j \frac{\partial }{\partial z_i} 1/r_{ij}=0
class Bodies
{
 public:
   
    std::vector<Body> data;
    int N; // the number of bodies
    MyInterval massN;

    /* Constructors */
  //  Bodies() {};

   Bodies(int num_bodies) ; /* creates structure with <num_bodies> bodies, we have to be carreful with this constructor because
             it does not initiate masses */

   
    #ifdef PLANAR
    Bodies(int num_bodies,bool addOverSize, vector<MyDouble> &xInit, vector<MyDouble> &yInit,vector<MyDouble>  &massInit); 
    #endif
    #ifdef SPATIAL
	 Bodies(int num_bodies,bool addOverSize, vector<MyDouble> &xInit, vector<MyDouble> &yInit, vector<MyDouble> &zInit,vector<MyDouble>  &massInit); 
    #endif 
	// creates from the data read from <xInit>,<yInit>,<zInit>, <massInit> which contain the input data regarding bodies 
	//  when addOverSize - then the initial data is a bit extended
   
    int dimRedConfSpace() const; 
	// returns the dimension of the reduced configuration space
 // in the planar case we drop:  y_{n-2}, x_{n-1},y_{n-1}
 // in the spatial case we drop: z_0,y_{n-2},z_{n-2}, x_{n-1},y_{n-1},z_{n-1}
   
    MyVector toReducedSpace() const;
	// returns the vector giving <this> in the reduced space coordinates
	
    // the functions return the position of the last body - which is computed from the center of mass condition
    MyInterval xN() const;
    MyInterval yN() const;
	#ifdef SPATIAL
	 MyInterval zN() const;
	 #endif

    /** Indexing bodies */

    inline MyInterval x(int i) const { return (i == this->N - 1) ? this->xN() : this->data[i].x; }
    inline MyInterval y(int i) const { return (i == this->N - 1) ? this->yN() : this->data[i].y; }
	#ifdef SPATIAL
	  inline MyInterval z(int i) const { return (i == this->N - 1) ? this->zN() : this->data[i].z; }
	#endif
    inline MyInterval mass(int i) const { return (i == this->N - 1) ? this->massN : this->data[i].mass; }

    /** Set the i-th body's x coord to val */
    void setX(int i, const MyInterval &val);
    /** Set the i-th body's y coord to val */
    void setY(int i, const MyInterval &val);
	
	#ifdef SPATIAL
	  void setZ(int i,const MyInterval &val);
    /** Set the i-th body's y mass to val */
	#endif
	
    /** Set the i-th body's y mass to val */
    void setMass(int i,const MyInterval &val);

 
    /** Radius */
    // r_{ij} = \sqrt{x_{ij}^2 + y_{ij}^2+z_{ij}^2}
    MyInterval radius(int i, int j)  const; 
	
	
    /** Derivatives */
    #ifdef PLANAR
	void derivatives(int i, int j, MyInterval &dpx, MyInterval &dpy,  MyInterval &ppxy) const;
    #endif	
	#ifdef SPATIAL
	void derivatives(int i, int j, MyInterval &dpx, MyInterval &dpy, MyInterval &dpz,
                     MyInterval &ppxy, MyInterval &ppxz, MyInterval &ppyz) const;
	#endif
	// it makes sense for i \neq j only
	// WARNING - the minus signs are important
//   dpx=-\frac{D^2 }{D x_i^2} 1/r_{ij}= \frac{D^2 }{D x_i Dx_j} 1/r_{ij}
//   dpy=-\frac{D^2 }{D y_i^2} 1/r_{ij}= \frac{D^2 }{Dy_i Dy_j} 1/r_{ij}	
//   dpz=-\frac{D^2 }{D z_i^2} 1/r_{ij}= \frac{D^2 }{D z_i D z_j} 1/r_{ij}
//   ppxy=\frac{D^2 }{Dx_i Dy_i} 1/r_{ij}=- \frac{D^2 }{D x_i y_j} 1/r_{ij}
//   ppxz=\frac{D^2 }{D x_i z_i} 1/r_{ij}=- \frac{D^2 }{D x_i z_j} 1/r_{ij}	
//   ppyz=\frac{D^2 }{D y_i z_i} 1/r_{ij}=- \frac{D^2 }{D y_i z_j} 1/r_{ij}					 
	// dpx = \frac{1}{r_{ij}^3} - 3\frac{(x_i - x_j)^2}{r_{ij}^5}
    // dpy = \frac{1}{r_{ij}^3} - 3\frac{(y_i - y_j)^2}{r_{ij}^5}
	// dpz = \frac{1}{r_{ij}^3} - 3\frac{(z_i - z_j)^2}{r_{ij}^5}
	// ppxy= 3 \frac{(x_i - x_j)(y_i - y_j)}{r_{ij}^5}
	// ppxz= 3 \frac{(x_i - x_j)(z_i - z_j)}{r_{ij}^5}
	// ppyz= 3 \frac{(y_i - y_j)(z_i - z_j)}{r_{ij}^5}
	

    /** Collision */
    bool collisionHappened() const;
// true is returned when positions of some bodies overlap


    /** Functions */
	
	 // gravity force:
	#ifdef PLANAR
bool GXY(int i, MyInterval &sumX, MyInterval &sumY) const;
#endif
#ifdef SPATIAL
bool GXYZ(int i, MyInterval &sumX, MyInterval &sumY, MyInterval &sumZ) const;
#endif
// computes xyz-coordinates of \sum_{j,j\neq i} \frac{m_j}{r_{ij}^3}(q_i - q_j)
// false is returned when the i-th body is in possible collision with other body, in such situation sumX,sumY,sumZ - contain some trash 
	
	#ifdef PLANAR
    bool fXY(int i, MyInterval &fXi, MyInterval &gXi, MyInterval &fYi, MyInterval &gYi) const;
    #endif
    #ifdef SPATIAL
    bool fXYZ(int i, MyInterval &fXi, MyInterval &gXi, MyInterval &fYi, MyInterval &gYi, MyInterval &fZi, MyInterval &gZi) const;
	#endif
	// computes xyz-coordinates of  f_{q_i}= q_i- G_i ,
//  G_i=\sum_{j,j\neq i} \frac{m_j}{r_{ij}^3}(q_i - q_j)
// false is returned when the i-th body is in possible collision with other body, in such situation fXi, gXi, fYi, gYi, fZi, gZi - contain some trash


    bool compFRed(MyVector &FRed) const ;
	// returns true if there is no colision and in such case fxi,fyi,fzi are put in FRed in a correct way
	// so that can be used in the Krawczyk or Newton operator

    /** Martix of derivatives of the reduced system */
    MyInterval DxfX(int i, int k, MyMatrix &dpxM) const; 
	// Calculate: df^x_i/dx_k   -taking into account  the fact that q_{n-1} is computed from the center of mass condition */
    // On input:
   // dpxM[i][j]= \frac{1}{r_{ij}^3} - 3\frac{(x_i - x_j)^2}{r_{ij}^5}=-\frac{D^2 }{D x_i^2} 1/r_{ij}= \frac{D^2 }{D x_i Dx_j} 1/r_{ij}, i,j=0,..,N-1

    MyInterval DyfY(int i, int k, MyMatrix &dpyM) const; /* df^y_i/dy_k  - -taking into account  the fact that q_{n-1} is computed from the center of mass condition*/
    MyInterval DxfY(int i, int k, MyMatrix &ppM) const;  /* df^x_i/dy_k = df^y_i/dx_k - -taking into account  the fact that q_{n-1} is computed from the center of mass condition */  //  ZOBACZYC CZY TO JEST NAPEWNO OK
    #ifdef SPATIAL
    MyInterval DzfZ(int i, int k, MyMatrix &dpzM) const;  /* df^z_i/dz_k -taking into account  the fact that q_{n-1} is computed from the center of mass condition */
    MyInterval DxfZ(int i, int k, MyMatrix &ppxzM) const; /* df_i^x/dz_k = df_i^z/dx_k -taking into account  the fact that q_{n-1} is computed from the center of mass condition */
    MyInterval DyfZ(int i, int k, MyMatrix &ppyzM) const; /* df_i^y/dz_k = df_i^z/dy_k  -taking into account  the fact that q_{n-1} is computed from the center of mass condition */
	#endif


    /** Utils */

    void printData(bool writeDistFromZero = false) const;
    void printDiam() const;
    bool writeCCformatMath(ostream &mathFile, ostream &mathListFile) const;
    void writeBinaryData(ostream &outFile) const;
    bool readBinaryData(istream &inFile);

 
   
	MyDouble maxDiam(int &max_i, int &coord) const;
/**  we return the diameter (rigorous upper bound) of the largest  of active coordinates - 
	 we do not test the (N-1) body, which is computed from the center of mass condition
	 and the varaibles set to 0
   On output:
      min_i - is an index of the body with the smallest interval  
   coord = 0 =>  x, coord = 1 => y, coord = 2 => z **/
  
    MyDouble minDiam(int &min_i, int &coord) const;
	/** we return the diameter of the smallest interval.
	 we do not test the (N-1) body, which is computed from the center of mass condition
	 and the varaibles set to 0
   On output:
      min_i - is an index of the body with the smallest interval  
   coord = 0 =>  x, coord = 1 => y, coord = 2 => z **/
   

    Bodies *calculateMidBodies() const;
	
};


void computeDFRedMatrix(const Bodies &bodies, MyMatrix &DF);
// On output:
//  DF - is the matrix of partial derivatives od fReduced with respect to all active variables



MyDouble volume(const Bodies &bodies);
//  returns volume of <bodies> in the reduced configuration space

void vectorIndicesToNumberOfBodies(int i, int &iB, int &nB);
// From vector of reduced coordinates index <i> 
//  on output: we return <nB> \in (0,1,2) indicating respectively (x,y,z)
//                       <iB> the index of body

int fromBodiesToVectorIndices(int iB, int nB);
// this is an inverse function to <vectorIndicesToNumberOfBodies>
//  given the body <iB> and variable <nb> \in {0,1,2} (meaning (x,y,z))
// the function returns the corresponding index in the vector cointaining the reduced configuration space

bool cubeSubSetInterior(const MyVector &Nx, const Bodies &bodies);
// returns true when in the reduced configuration spacd Nx \subset bodies



bool cubeSuperSet(const MyVector Nx,const  Bodies &bodies);
// returns true iff in the reduced configuration space  bodies \subset Nx

bool cubeEmptyIntersection(const MyVector &Nx, const Bodies &bodies);
//  returns true iff Nx \cap reduced_coord(bodies) = \emptyset

MyVector cubeIntersection(const MyVector Nx,const Bodies &bodies);
//  returns vector in the reduced space equal to Nx \cap Reduced(bodies)
// it should be checked before that this intersection is non-empty

void updateBodies(const MyVector &Nx, Bodies &bodies);
//  function does bodies=Nx where Nx is in the reduced configuration space

MyInterval potential(const Bodies &bodies);
// returns U = \sum_{i < j} \frac{m_i*m_j}{r_{i, j}} 
//   it is assumed that no collision happens (this should be check before calling this function)

MyInterval momentInertia(const Bodies &bodies);
// returns I = \sum_{i} m_i*|q_i|^2 

void writePotAndInertia(const Bodies &bodies,bool writeMoeckelPot);
//  it makes to have writeMoeckelPot=true when masses are equal in 2D
//   otherwise we suggest false

void intervalHullBodies(const Bodies &sol1,const Bodies &sol2, Bodies &unionBodies);
// On output: unionBodies=IntervalHull(sol1,sol2)

void calculatePuffedBodies(const Bodies &bodies, MyInterval addedInterval, Bodies &puffedBodies);
/** "Puffs" the box of each body - by adding addedInterval, to each active variable in the reduced configuration space 
it assumed that <puffedBodies> has the same masses as <bodies>, so we do not copy the masses */

int readBodiesFromBinaryFile(int numBodies, const char *nazwaPliku, vector<Bodies> &ccSolutions);
//  reads bodies from the binary file <nazwaPliku> and places them in ccSolutions
//   returns the number of bodies read

void permuteMassesOfBodies(Bodies &bodies, int perm[]);
 // on output:
    // mass(bodies(i)) = mass(Bodies(input)(perm[i])) 
 // we do not change anything on coordinates, these must be set by some other function	

#endif // ifndef bodies_h
